/*
(c) Copyright Taiichi Yuasa and Masami Hagiya, 1984.  All rights reserved.
Copying of this file is authorized to users who have executed the true and
proper "License Agreement for Kyoto Common LISP" with SIGLISP.
*/
/*
	bitop.c

	Sets and/or tests bits in the mark table.  The first 8 words of the
	Lisp data area have their mark bits in the byte location pointed to
	by MARK_TABLE, and the next 8 words in the next byte location,
	and so on.

	GET_MARK_BIT(X)
	Returns 1 if the mark bit for the word address X is on.  Otherwise,
	returns 0.

	SET_MARK_BIT(X)
	Sets on the mark bit for the word address X.

	GET_SET_MARK_BIT(X)
	Sets on the mark bit for the word address X, and returns 1 if the
	mark bit was previously on.  Returns 0 otherwise.

*/
#include "include.h"

#ifdef VAX

get_mark_bit(x)
int *x;
{
	asm("	ashl	$-2,4(ap),r2");
	asm("	ashl	$-3,r2,r3");
	asm("	bicl2	$-8,r2");
	asm("	clrl	r0");
	asm("	addl2	_mark_table,r3");
	asm("	bbc	r2,(r3),1f");
	asm("	incl	r0");
	asm("1:");
}

set_mark_bit(x)
int *x;
{
	asm("	ashl	$-2,4(ap),r2");
	asm("	ashl	$-3,r2,r3");
	asm("	bicl2	$-8,r2");
	asm("	addl2	_mark_table,r3");
	asm("	bbcs	r2,(r3),1f");
	asm("1:");
}

get_set_mark_bit(x)
int *x;
{
	asm("	ashl	$-2,4(ap),r2");
	asm("	ashl	$-3,r2,r3");
	asm("	bicl2	$-8,r2");
	asm("	clrl	r0");
	asm("	addl2	_mark_table,r3");
	asm("	bbcs	r2,(r3),1f");
	asm("	incl	r0");
	asm("1:");
}

#endif

#ifdef MC68K

get_mark_bit(x)
int *x;
{
	asm("	movl	a6@(8),d0");
	asm("	lsrl	#2,d0");
	asm("	movl	d0,d1");
	asm("	lsrl	#3,d1");
	asm("	movl	_mark_table,a0");
	asm("	btst	d0,a0@(0,d1:L)");
	asm("	sne	d0");
	asm("	andl	#1,d0");
}

set_mark_bit(x)
int *x;
{
	asm("	movl	a6@(8),d0");
	asm("	lsrl	#2,d0");
	asm("	movl	d0,d1");
	asm("	lsrl	#3,d1");
	asm("	movl	_mark_table,a0");
	asm("	bset	d0,a0@(0,d1:L)");
}

get_set_mark_bit(x)
int *x;
{
	asm("	movl	a6@(8),d0");
	asm("	lsrl	#2,d0");
	asm("	movl	d0,d1");
	asm("	lsrl	#3,d1");
	asm("	movl	_mark_table,a0");
	asm("	bset	d0,a0@(0,d1:L)");
	asm("	sne	d0");
	asm("	andl	#1,d0");
}

#endif

#ifdef SUN3

get_mark_bit(x)
int *x;
{
	asm("	movl	a6@(8),d0");
	asm("	lsrl	#2,d0");
	asm("	movl	d0,d1");
	asm("	lsrl	#3,d1");
	asm("	movl	_mark_table,a0");
	asm("	btst	d0,a0@(0,d1:L)");
	asm("	sne	d0");
	asm("	andl	#1,d0");
}

set_mark_bit(x)
int *x;
{
	asm("	movl	a6@(8),d0");
	asm("	lsrl	#2,d0");
	asm("	movl	d0,d1");
	asm("	lsrl	#3,d1");
	asm("	movl	_mark_table,a0");
	asm("	bset	d0,a0@(0,d1:L)");
}

get_set_mark_bit(x)
int *x;
{
	asm("	movl	a6@(8),d0");
	asm("	lsrl	#2,d0");
	asm("	movl	d0,d1");
	asm("	lsrl	#3,d1");
	asm("	movl	_mark_table,a0");
	asm("	bset	d0,a0@(0,d1:L)");
	asm("	sne	d0");
	asm("	andl	#1,d0");
}

#endif

#ifdef ATT3B2

get_mark_bit(x)
int *x;
{
	asm("	subw3	&0x80800000,0(%ap),%r1"); /* R1 <= address offset */
	asm("	lrsw3	&2,%r1,%r1");		/* R1 <= bit offset */
	asm("	lrsw3	&3,%r1,%r0");		/* R0 <= byte offset */
	asm("	andw2	&0x7,%r1");
	asm("	addw2	mark_table, %r0");
	asm("	movb	0(%r0),%r0");		/* R0 <= table entry */
	asm("	lrsw3	%r1,%r0,%r0");
	asm("	andw2	&1,%r0");		/* LSB(R0) = the bit */
}

set_mark_bit(x)
int *x;
{
	asm("	subw3	&0x80800000,0(%ap),%r0"); /* R0 <= address offset */
	asm("	lrsw3	&2,%r0,%r0");		/* R0 <= bit offset */
	asm("	lrsw3	&3,%r0,%r1");		/* R1 <= byte offset */
	asm("	andw2	&0x7,%r0");
	asm("	llsw3	%r0,&1,%r0");		/* R0 <= mask */
	asm("	addw2	mark_table,%r1");
	asm("	orb2	%r0,0(%r1)");
}

get_set_mark_bit(x)
int *x;
{
	register int *r8;

	asm("	subw3	&0x80800000,0(%ap),%r1"); /* R1 <= address offset */
	asm("	lrsw3	&2,%r1,%r1");		/* R1 <= bit offset */
	asm("	lrsw3	&3,%r1,%r2");		/* R2 <= byte offset */
	asm("	andw2	&0x7,%r1");
	asm("	addw2	mark_table,%r2");
	asm("	movb	0(%r2),%r0");		/* R0 <= table entry */
	asm("	llsw3	%r1,&1,%r8");		/* R8 <= mask */ 
	asm("	orb2	%r8,0(%r2)");		/* set */
	asm("	lrsw3	%r1,%r0,%r0");
	asm("	andw2	&1,%r0");		/* get */
}

#endif

#ifdef NS32K

get_mark_bit(x)
int *x;
{
	asm("	movqd	0,r0");
	asm("	movd	8(fp),r1");
	asm("	ashd	-2,r1");
	asm("	movd	_mark_table,r2");
	asm("	tbitd	r1,0(r2)");
	asm("	bfc	.t1");
	asm("	movqd	1,r0");
	asm(".t1:");
}

set_mark_bit(x)
int *x;
{
	asm("	movd	8(fp),r1");
	asm("	ashd	-2,r1");
	asm("	movd	_mark_table,r2");
	asm("	sbitd	r1,0(r2)");
}

get_set_mark_bit(x)
int *x;
{
	asm("	movqd	0,r0");
	asm("	movd	8(fp),r1");
	asm("	ashd	-2,r1");
	asm("	movd	_mark_table,r2");
	asm("	sbitd	r1,0(r2)");
	asm("	bfc	.t2");
	asm("	movqd	1,r0");
	asm(".t2:");
}

#endif

#ifdef BITOP_C

extern int *mark_table;

get_mark_bit(x)
int x;
{
	int y;

	y = (*(mark_table+(x/4/32)) >> (x/4%32)) & 1;
	return(y);
}

set_mark_bit(x)
int x;
{
	int y;

	y = 1 << (x/4%32);
	y = (*(mark_table+(x/4/32))) | y;
	*(mark_table+ (x/4/32))=y;
}

get_set_mark_bit(x)
int x;
{
	int y;

	y = get_mark_bit(x);
	set_mark_bit(x);
	return(y);
}

#endif

/* IBMRT */

#ifdef NEWS

get_mark_bit(x)
int *x;
{
	asm("	move.l	(8,fp),d0");
	asm("	lsr.l	#2,d0");
	asm("	move.l	d0,d1");
	asm("	lsr.l	#3,d1");
	asm("	move.l	_mark_table,a0");
	asm("	btst	d0,0(a0,d1.l)");
	asm("	sne	d0");
	asm("	and.l	#1,d0");
}

set_mark_bit(x)
int *x;
{
	asm("	move.l	(8,fp),d0");
	asm("	lsr.l	#2,d0");
	asm("	move.l	d0,d1");
	asm("	lsr.l	#3,d1");
	asm("	move.l	_mark_table,a0");
	asm("	bset	d0,0(a0,d1.l)");
}

get_set_mark_bit(x)
int *x;
{
	asm("	move.l	(8,fp),d0");
	asm("	lsr.l	#2,d0");
	asm("	move.l	d0,d1");
	asm("	lsr.l	#3,d1");
	asm("	move.l	_mark_table,a0");
	asm("	bset	d0,0(a0,d1.l)");
	asm("	sne	d0");
	asm("	and.l	#1,d0");
}

#endif

#ifdef TAHOE

get_mark_bit(x)
int *x;
{
       asm("   shrl   $2,4(fp),r2");
       asm("   shrl   $3,r2,r3");
       asm("   andl2  $7,r2");
       asm("   clrl   r0");
       asm("   addl2  _mark_table,r3");
       asm("   movb   (r3),r4"); 
       asm("   bbc    r2,r4,1f");
       asm("   incl   r0");
       asm("1: ");
}
        
set_mark_bit(x)
int *x;
{
       asm("   shrl   $2,4(fp),r2");
       asm("   shrl   $3,r2,r3");
       asm("   andl2  $7,r2");
       asm("   movb   $1,r4");
       asm("   shll   r2,r4,r4");
       asm("   addl2  _mark_table,r3");
       asm("   orb2   r4,(r3)");
}
   
get_set_mark_bit(x)
int *x;
{
       asm("   shrl   $2,4(fp),r2");     /* r2 := word offset */
       asm("   shrl   $3,r2,r3");        /* r3 := byte offset in mark-table */
       asm("   andl2  $7,r2");           /* r2 := bit offset in mark-byte */
       asm("   addl2  _mark_table,r3");
       asm("   movb   (r3),r4");         /* r4 := byte adr in mark-table */
       asm("   clrl   r0");              /* r0 initialized to bit clear */
       asm("   bbc    r2,r4,1f");        /* bit clear? */
       asm("   incl   r0");              /* yes. set r0 to 1 */
       asm("1: movb   $1,r4");           /* Change bit offset to mask bit */
       asm("   shll   r2,r4,r4");        /* and set bit on mark-table */
       asm("   orb2   r4,(r3)");
}             
 
#endif

#ifdef OMRON

get_mark_bit(x)
int *x;
{
	asm("	mov.l	(8,%fp),%d0");
	asm("	lsr.l	&2,%d0");
	asm("	mov.l	%d0,%d1");
	asm("	lsr.l	&3,%d1");
	asm("	mov.l	mark_table,%a0");
	asm("	btst	%d0,(%a0,%d1.l*1)");
	asm("	sne.b	%d0");
	asm("	and.l	&1,%d0");
}

set_mark_bit(x)
int *x;
{
	asm("	mov.l	(8,%fp),%d0");
	asm("	lsr.l	&2,%d0");
	asm("	mov.l	%d0,%d1");
	asm("	lsr.l	&3,%d1");
	asm("	mov.l	mark_table,%a0");
	asm("	bset	%d0,(%a0,%d1.l*1)");
}

get_set_mark_bit(x)
int *x;
{
	asm("	mov.l	(8,%fp),%d0");
	asm("	lsr.l	&2,%d0");
	asm("	mov.l	%d0,%d1");
	asm("	lsr.l	&3,%d1");
	asm("	mov.l	mark_table,%a0");
	asm("	bset	%d0,(%a0,%d1.l*1)");
	asm("	sne.b	%d0");
	asm("	and.l	&1,%d0");
}

#endif
