#include <stdio.h>

/*
	input file format
*/

struct nsection {
	unsigned int logical_start;
	unsigned int logical_length;
	unsigned int file_start;
	unsigned int file_length;
	unsigned int file_reloc;
	unsigned int file_reloc_length;
	unsigned int attributes;
	};

struct prformat {
	unsigned short magic_number;       /*  magic                 */
	unsigned short format_rev_number;  /*  format rev. number    */
	unsigned short num_non_exec_sect;  /*  # of non-exec.  sects */
	unsigned short num_exec_sect;      /*  # of executable sects */
	struct nsection syshead;        /*  system header */
	struct nsection symbolt;        /*  symbol table  */
	struct nsection ds;             /*                */
	struct nsection dl;             /*                */
	struct nsection lt;             /*                */
	struct nsection exec[2];        /*  two executable sections */
	} head;

struct part_sym {
	unsigned short	p_type;
	unsigned short	p_value_size;
	unsigned int	p_value;
	unsigned int	p_wleng;
	unsigned short	p_name_len;
};

struct other_sym {
	unsigned short	o_type;
	unsigned short	o_value_size;
	unsigned int	o_value;
	unsigned short	o_name_len;
};

union sym_ent {
	struct other_sym o;
	struct part_sym  p;
} symbol;

char	symbol_name[256];

/*
	for new symbol table
*/

#define PAGESIZE 2048

struct page_header {
	unsigned short count;		/* # of entries in this page */
	unsigned short npage;		/* next page no of the same hash */
};

struct symbol_entry {
	int		value;		/* symbol value */
	unsigned short	blength;	/* symbol name byte length */
};	


/*
	tables
*/

#define HSIZE 8

struct {
	int	remain;
	struct page_header *headerp;
	struct symbol_entry *entryp;
} htable[HSIZE];

struct pageq {
	char	*pagep;
	struct pageq *npage;
} *page_start, *cpage;

int	page_count = 0;

FILE *in;
FILE *out;

alloc_newpage(h)
int h;
{
	char	*p, *alloc();
	struct pageq *q;

	p = alloc(PAGESIZE);
	q = (struct pageq *)alloc(sizeof(struct pageq));
	cpage->npage = q;
	cpage = q;
	q->pagep = p;
	q->npage = 0;

	htable[h].headerp->npage = page_count;

	htable[h].headerp = (struct page_header *)p;
	htable[h].headerp->count = 0;
	htable[h].headerp->npage = 0;
	htable[h].remain = PAGESIZE - sizeof(struct page_header);
	htable[h].entryp = (struct symbol_entry *)(htable[h].headerp + 1);

	page_count++;
}

init_page()
{
	int	h;
	char	*p, *alloc();
	struct pageq *q;

	p = alloc(PAGESIZE);
	page_start = cpage = q =
		(struct pageq *)alloc(sizeof(struct pageq));
	q->pagep = p;
	q->npage = 0;

	htable[0].headerp = (struct page_header *)p;
	htable[0].headerp->count = 0;
	htable[0].headerp->npage = 0;
	htable[0].remain = PAGESIZE - sizeof(struct page_header);
	htable[0].entryp = (struct symbol_entry *)(htable[0].headerp + 1);

	page_count++;

	for (h = 1; h < HSIZE; h++) {
		p = alloc(PAGESIZE);
		q = (struct pageq *)alloc(sizeof(struct pageq));
		cpage->npage = q;
		cpage = q;
		q->pagep = p;
		q->npage = 0;

		htable[h].headerp = (struct page_header *)p;
		htable[h].headerp->count = 0;
		htable[h].headerp->npage = 0;
		htable[h].remain = PAGESIZE - sizeof(struct page_header);
		htable[h].entryp =
			(struct symbol_entry *)(htable[h].headerp + 1);

		page_count++;
	}
}

int
hash(name, len)
char *name;
int len;
{
	return((name[0] + name[len-1]) % HSIZE);
}

void
add_symbol(name, len, type, val)
char *name;
int len;
int val;
{
	int	h;

	h = hash(name, len);

	if (htable[h].remain < len + 3) alloc_newpage(h);

	(htable[h].headerp->count)++;

/*	htable[h].entryp->type = type;	*/
	htable[h].entryp->value = val;
	htable[h].entryp->blength = len;
	strncpy((char *)(htable[h].entryp + 1), name, len);

	htable[h].entryp =
		(struct symbol_entry *)((char *)(htable[h].entryp + 1)
			+ len + (len & 1));
	htable[h].remain -= (sizeof(struct symbol_entry) + len + (len & 1));
}

void
do_one_symbol()
{
	int	len, blen;
	int	pos;
	int	type, val;

	pos = ftell(in);

	fread((char *)&symbol, sizeof(struct part_sym), 1, in);
	if (ferror(in)) {
		printf("read error %o", lasterror());
		exit(1);
	}

	if ((len = symbol.p.p_value_size) == 4) {
		type = symbol.o.o_type;
		val = symbol.o.o_value;
		blen = symbol.o.o_name_len;
	} else {
		type = symbol.p.p_type;
		val = symbol.p.p_value;
		blen = symbol.p.p_name_len;
	}

	if (fseek(in, pos + len + 6, 0) < 0) {
		printf("seek error %o\n", lasterror());
		exit(1);
	}

	fread(symbol_name, blen + (blen & 1), 1, in);

	add_symbol(symbol_name, blen, type, val);
}

main(argc, argv)
int argc;
char *argv[];
{
	unsigned int	ststart;
	unsigned int	stlen;
	unsigned int	stcount;
	int	i;
	struct pageq	*q;

	if (argc < 3) {
		printf("Usage: symboltb input output\n");
		exit(1);
	}

	if ((in = fopen(argv[1], "r")) == NULL) {
		printf("can't open %s %o\n", argv[1], lasterror());
		exit(1);
	}

	if ((out = fopen(argv[2], "w+")) == NULL) {
		printf("can't open %s %o\n", argv[2], lasterror());
		exit(1);
	}

	fread((char *)(&head), sizeof(struct prformat), 1, in);
	if (ferror(in)) {
		printf("read error %o\n", lasterror());
		exit(1);
	}

	ststart = head.symbolt.file_start;
	stcount = head.symbolt.logical_length;

	fseek(in, ststart, 0);
	if (ferror(in)) {
		printf("seek error %o\n", lasterror());
		exit(1);
	}

	init_page();

	while (stcount-- > 0) do_one_symbol();

	for (q = page_start; q != 0; q = q->npage) {
		fwrite(q->pagep, PAGESIZE, 1, out);
		if (ferror(out)) {
			printf("write error %o", lasterror());
			exit(1);
		}
	}

	fclose(in);
	fclose(out);
}
